import re
import os
import itertools
import logging
import argparse

log = logging.getLogger(__name__)


def terminate(process):
    try:
        process.terminate()
    except WindowsError:
        pass  # process has probably already terminated on its own


class consecutive_count:
    def __init__(self):
        self.count = 0
        self.last = None

    def __call__(self, item):
        if item == self.last:
            self.count += 1
        else:
            self.count = 0
            self.last = item
        return self.count


def consecutive_same(items):
    """A generator that returns the count of consecutive
    preceeding items that match the current item.
    >>> tuple(consecutive_same([1,1,2,2,3,3,3]))
    (0, 1, 0, 1, 0, 1, 2)

    Note this will trivially produce zeros for most
    inputs.
    >>> tuple(consecutive_same([1,2,3]))
    (0, 0, 0)
    """
    counter = consecutive_count()
    return map(counter, items)


def within_consecutive_limit(limit):
    counter = consecutive_count()

    def within_limit(item):
        return counter(item) < limit

    return within_limit


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('-t', '--title', help="The title to use", default="")
    parser.add_argument('device', nargs='?')
    args = parser.parse_args()
    title = args.title
    device = args.device
    log.info(['using default device', 'using device %s' % device][bool(device)])
    log.info(['using default title', 'using title %s' % title][bool(title)])
    return device, title


def build_command(dvd_device=None, title=""):
    from jaraco.media.dvd import MEncoderCommand, HyphenArgs

    command = MEncoderCommand()
    command.source = ['dvd://%(title)s' % vars()]
    command.audio_options = HyphenArgs(nosound=None)
    limit = 24
    round = 16
    command.video_filter = HyphenArgs(vf='cropdetect=%(limit)s:%(round)s' % locals())
    command.video_options = HyphenArgs(ovc='lavc')
    command['o'] = os.path.devnull
    # command['ss'] = '20' # skip the first 20 seconds
    # As a rule-of-thumb, use the 2nd or 3rd chapter to determine the crop;
    #  The first chapter can tend to have a different format.
    command['chapter'] = '2'

    dvd_device and command.set_device(dvd_device)
    return command


def get_input(command=None):
    from subprocess import Popen, PIPE, list2cmdline

    null = open('NUL', 'w')
    command = command or build_command(*parse_args())
    args = tuple(command.get_args())
    log.debug(list2cmdline(args))
    mencoder = Popen(args, stdout=PIPE, stderr=null)
    return mencoder


class InsufficientFramesError(RuntimeError):
    pass


def process_input(process, n_frames=1000):
    pattern = re.compile(r'.*crop=(\d+:\d+:\d+:\d+).*')

    crop_matches = filter(None, map(pattern.match, process.stdout))
    crop_values = map(lambda match: match.group(1), crop_matches)
    preceeding_items = itertools.takewhile(
        within_consecutive_limit(n_frames), crop_values
    )

    log.info('processed %d frames', len(tuple(preceeding_items)))
    try:
        target = crop_values.next()
    except StopIteration:
        raise InsufficientFramesError(
            "Not enough frames to detect %d consecutive same" % n_frames
        )
    finally:
        clean_up(process)
    return target


def clean_up(process):
    terminate(process)
    process.stdout.flush()


def get_crop(dvd_device=None, title=''):
    log.info('Detecting crop')
    return process_input(get_input(build_command(dvd_device, title)))


def execute(command=None):
    logging.basicConfig(level=logging.INFO)
    mencoder = get_input(command)
    try:
        log.info('crop is %s', process_input(mencoder))
    except InsufficientFramesError as e:
        log.warning(e)


if __name__ == '__main__':
    execute()



#!/usr/bin/python
# Little script to depack Matroska file, and repack them
# in an MP4 format.

import os
import argparse
import subprocess


def message(msg):
    print("=" * 78)
    print("=", msg)
    print("=" * 78)


def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('filename')
    return parser.parse_args()


def handle_command_line():
    args = get_args()
    filename = args.filename
    basename = os.path.basename(filename)
    name, ext = os.path.splitext(basename)
    message("Unpacking file: %s" % filename)
    subprocess.Popen(
        [
            r'C:\Program Files (x86)\mkvtoolnix\mkvextract',  # 'mkvextract',
            'tracks',
            filename,
            '2:temp_video.mp4',
            '1:temp_audio.ogg',
            # f'3:{name}.srt',
        ]
    ).wait()

    message(f"Repacking file: {name}.mp4")
    subprocess.Popen(
        [
            r'C:\Program Files\ffmpeg-git-9251942-win64-shared\bin\ffmpeg.exe',
            # 'ffmpeg',
            '-i',
            'temp_audio.ogg',
            '-i',
            'temp_video.mp4',
            '-vcodec',
            'copy',
            # 'libx264',
            '-r',
            '47.95',
            name + '.mp4',
        ]
    ).wait()

    message("Cleaning files")
    map(os.remove, ['temp_video.mp4', 'temp_audio.ogg'])


if __name__ == "__main__":
    handle_command_line()



import itertools
import os
import urllib
import traceback

import path
import cherrypy
import genshi.template
import httpagentparser

from . import config


class Movie:
    def __init__(self, filename, class_):
        self.class_ = class_
        self.filename = filename

    @property
    def title(self):
        filename = os.path.basename(self.filename)
        return os.path.splitext(filename)[0]

    @property
    def link(self):
        return 'http://www.imdb.com/find?s=all&' + self.search

    @property
    def search(self):
        q = urllib.urlencode(dict(q=self.title))
        return q


class Index:
    def __init__(self, filter=None):
        self.filter = filter

    def get_named_map(self):
        result = {dir.name: dir for dir in config.movies_root.dirs()}
        result[path.Path('')] = config.movies_root
        return result

    def get_media(self, root):
        items = (os.path.join(root, item) for item in os.listdir(root))
        files = itertools.ifilter(os.path.isfile, items)
        mfiles = itertools.ifilter(self.is_media, files)
        return mfiles

    def __iter__(self):
        return itertools.ifilter(self.filter, self.get_all())

    def get_all(self):
        for class_, root in self.get_named_map().items():
            for media in self.get_media(root):
                yield Movie(media, class_)

    @staticmethod
    def is_media(path):
        media_extensions = ('.mp4', '.avi')
        name, ext = os.path.splitext(path)
        return ext in media_extensions


class Site:
    @cherrypy.expose
    def index(self):
        cherrypy.response.headers['Content-Type'] = 'application/xhtml+xml'
        template = self.get_template()

        def not_watched(m):
            return m.class_ != 'watched'

        movies = Index(filter=not_watched)
        res = template.generate(movies=movies, title="Movies to Watch")
        return res.render('xml')

    def get_template(self):
        """
        iPhone user agent is
        Mozilla/5.0 (iPhone; U; CPU iPhone OS 3_1_2 like Mac OS X; en-us) \
        AppleWebKit/528.18 (KHTML, like Gecko) Version/4.0 Mobile/7D11 Safari/528.16
        """
        agent = httpagentparser.detect(cherrypy.request.headers['User-Agent'])
        is_ios = agent.get('dist', {}).get('name', None) in ['iPhone', 'IPad']
        template_type = 'iweb' if is_ios else ''
        template_text = globals().get(template_type + '_template', default_template)
        template = genshi.template.MarkupTemplate(template_text)
        return template

    @classmethod
    def setup_application(cls, root):
        config = {
            'global': {
                'tools.encode.on': True,
                'tools.encode.encoding': 'utf-8',
                'tools.decode.on': True,
                'tools.trailing_slash.on': True,
            },
        }

        return cherrypy.tree.mount(cls(), root, config)

    @classmethod
    def factory(cls):
        "The entry point for when the ISAPIDLL is triggered"
        try:
            import isapi_wsgi

            return isapi_wsgi.ISAPISimpleHandler(cls.setup_application('/media'))
        except Exception:
            traceback.print_exc()


__ExtensionFactory__ = Site.factory


def handle_isapi():
    "Install or remove the extension to the virtual directory"
    import isapi.install

    params = isapi.install.ISAPIParameters()
    # Setup the virtual directories - this is a list of directories our
    # extension uses - in this case only 1.
    # Each extension has a "script map" - this is the mapping of ISAPI
    # extensions.
    sm = [isapi.install.ScriptMapParams(Extension="*", Flags=0)]
    vd = isapi.install.VirtualDirParameters(
        Server="Default Web Site",
        Name="/media",
        Description="Media Index",
        ScriptMaps=sm,
        ScriptMapUpdate="end",
    )
    params.VirtualDirs = [vd]
    isapi.install.HandleCommandLine(params)


def serve():
    Site.setup_application('/')
    cherrypy.config.update(
        {
            'server.socket_host': '::0',
        }
    )
    if hasattr(cherrypy.engine, "signal_handler"):
        cherrypy.engine.signal_handler.subscribe()
    if hasattr(cherrypy.engine, "console_control_handler"):
        cherrypy.engine.console_control_handler.subscribe()
    cherrypy.engine.start()
    cherrypy.engine.block()


default_template = """\
<html xmlns="http://www.w3.org/1999/xhtml" \
xmlns:py="http://genshi.edgewall.org/">
    <body>
        <h1 py:content="title">Title</h1>
        <div py:for="movie in movies">
            <a href="${movie.link}" py:content="movie.title">Movie Title</a>
        </div>
    </body>
</html>
"""

iweb_template = """<!DOCTYPE html PUBLIC \
"-//W3C//DTD XHTML 1.0 Strict//EN" \
"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html
    xmlns="http://www.w3.org/1999/xhtml"
    xmlns:py="http://genshi.edgewall.org/">

<head>
<meta content="yes" name="apple-mobile-web-app-capable" />
<meta content="text/html; charset=iso-8859-1" http-equiv="Content-Type" />
<meta content="minimum-scale=1.0, width=device-width, maximum-scale=0.6667, \
user-scalable=no" name="viewport" />
<link href="/iweb/css/style.css" rel="stylesheet" media="screen" \
type="text/css" />
<script src="/iweb/javascript/functions.js" type="text/javascript"></script>
<title>Title of your page</title>
<meta content="keyword1,keyword2,keyword3" name="keywords" />
<meta content="Description of your page" name="description" />
</head>

<body>
<div id="topbar">
    <div id="title" py:content="title">Title</div>
</div>
<div id="content">

<ul py:for="movie in movies" class="pageitem">
    <li class="menu">
        <a href="imdb:///find?${movie.search}">
            <img alt="Description" src="/iweb/thumbs/video.png" />
            <span class="name">
                <span py:replace="movie.title">Movie Title</span>
                <span class="class" py:if="movie.class_">(\
<span py:replace="movie.class_" />)</span>
            </span>
            <span class="arrow"></span>
        </a>
    </li>
</ul>

</div>
<div id="footer">
    <!-- Support iWebKit by sending us traffic; please keep this footer on your \
page, consider it a thank you for our work :-) -->
    <a class="noeffect" href="http://iwebkit.net">Powered by iWebKit</a></div>

</body>

</html>
"""



import argparse
import re
import os
import subprocess
from os.path import join
from copy import deepcopy
import logging
import platform

import path
import inflect
from more_itertools import flatten

try:
    import win32api
except Exception:
    pass
from jaraco.text import trim, WordSet, FoldedCase

from jaraco.media import cropdetect
from jaraco.media.arguments import HyphenArgs, ColonDelimitedArgs

log = logging.getLogger(__name__)


rangePattern = re.compile(r'(\d+)(?:-(\d+))?')
delimiterPattern = re.compile(r'\s*[, ;]\s*')


def guess_output_filename(name):
    """
    >>> print(guess_output_filename('JEAN_DE_FLORETTE'))
    Jean De Florette

    >>> str(guess_output_filename(''))
    ''

    >>> str(guess_output_filename('COWBOYS_ALIENS_RENTAL'))
    'Cowboys Aliens'
    """
    rental = FoldedCase('rental')
    return WordSet.parse(name).trim_right(rental).capitalized().space_separated()


def get_source():
    default_device = path.Path('D:\\')
    if platform.system() == 'Linux':
        media = path.Path('/media')
        default_device = media.dirs()[0]
    if platform.system() == 'Darwin':
        volumes = path.Path('/Volumes')
        candidates = (
            subdir
            for subdir in volumes.dirs()
            if (subdir / 'BDMV').isdir() or (subdir / 'VIDEO_TS').isdir()
        )
        default_device = next(candidates, None)
    return os.environ.get('DVD', default_device)


def infer_name(device=None):
    device = device or get_source()
    try:
        label = win32api.GetVolumeInformation(device)[0]
    except Exception:
        label = os.path.basename(device)
    return guess_output_filename(label)


class MEncoderCommand:
    """
    >>> monkeypatch = getfixture('monkeypatch')
    >>> monkeypatch.setattr(
    ...     MEncoderCommand, 'find_mencoder_exe', lambda self: 'mencoder')
    >>> cmd = MEncoderCommand()
    >>> cmd.source = ['dvd://']
    >>> lavcopts = ColonDelimitedArgs(
    ...     vcodec='libx264',threads='2',vbitrate='1200',autoaspect=None,)
    >>> cmd.video_options = HyphenArgs(lavcopts=lavcopts)
    >>> cmd2 = cmd.copy()
    >>> cmd_args = tuple(cmd.get_args())
    >>> cmd2_args = tuple(cmd2.get_args())
    >>> assert cmd_args == cmd2_args, '%s != %s' % (cmd_args, cmd2_args)
    """

    def __init__(self):
        self.exe_path = [self.find_mencoder_exe()]
        self.other_options = HyphenArgs()

    @staticmethod
    def find_mencoder_exe():
        program_files = os.environ.get('PROGRAMFILES(X86)') or os.environ.get(
            'PROGRAMFILES'
        )
        search_path = [
            os.path.join(program_files, r'mencoder\mencoder.exe'),
        ]
        found = (path for path in search_path if os.path.exists(path))
        try:
            return next(found)
        except StopIteration:
            raise RuntimeError("Cannot find mencoder; admittedly didn't try very hard.")

    def copy(self):
        result = MEncoderCommand()
        # we need to do a deep copy so we make copies of all the args
        result.__dict__.update(deepcopy(self.__dict__))
        return result

    def get_args(self):
        arg_order = (
            'exe_path',
            'source',
            'device',
            'video_filter',
            'video_options',
            'audio_options',
            'other_options',
        )
        assert getattr(self, 'source', None) is not None
        for arg in arg_order:
            arg = getattr(self, arg, None)
            if arg is None:
                continue
            for value in arg:
                yield str(value)

    def set_device(self, value):
        assert os.path.exists(value), "Couldn't find device %s" % value
        self.device = HyphenArgs({'dvd-device': value})

    def __setitem__(self, key, value):
        self.other_options[key] = value

    def __getitem__(self, key):
        return self.other_options[key]


def expandRange(title_range):
    start, stop = rangePattern.match(title_range)
    stop = stop or start
    return range(int(start), int(stop) + 1)


def getTitles(title_spec_string):
    title_specs = delimiterPattern.split(title_spec_string)
    title_specs = flatten(map(expandRange, title_specs))


class MultiPassHandler:
    def __init__(self, command, passes=2):
        self.command = command.copy()
        self.setup_log_file()
        self.passes = passes

    def setup_log_file(self):
        filename, ext = os.path.splitext(self.command.other_options['o'])
        multi_pass_temp_file = filename + '_pass.log'
        self.command['passlogfile'] = multi_pass_temp_file

    def __iter__(self):
        inflect_eng = inflect.engine()
        for current_pass in range(1, self.passes + 1):
            current_pass_th = inflect_eng.ordinal(current_pass)
            log.info('Modifying command for %s pass.' % current_pass_th)
            method = [self.early_pass, self.last_pass][current_pass == self.passes]
            yield method(current_pass)

    def early_pass(self, pass_number):
        command = self.command.copy()
        command.audio_options = HyphenArgs(nosound=None)
        command.video_options['lavcopts'].update(turbo=None, vpass=str(pass_number))
        command['o'] = os.path.devnull
        return command

    def last_pass(self, pass_number):
        command = self.command.copy()
        command.video_options['lavcopts'].update(vpass=str(pass_number))
        return command

    def __del__(self):
        self.cleanup_log_file()

    def cleanup_log_file(self):
        try:
            filename = self.command['passlogfile']
        except KeyError:
            return
        try:
            os.path.exists(filename) and os.remove(filename)
        except Exception:
            log.error('Error removing logfile %s', filename)


def get_x264_options():
    lavcopts = ColonDelimitedArgs()
    lavcopts.update(vcodec='libx264')
    lavcopts.update(threads='2')
    lavcopts.update(vbitrate='1200')
    lavcopts.update(autoaspect=None)
    options = HyphenArgs()
    options.update(ovc='lavc')
    options.update(lavcopts=lavcopts)
    return options


def get_mpeg4_options():
    lavcopts = ColonDelimitedArgs()
    lavcopts.update(vcodec='mpeg4')
    lavcopts.update(vhq=None)
    lavcopts.update(vbitrate='1200')
    lavcopts.update(autoaspect=None)
    options = HyphenArgs()
    options.update(ovc='lavc')
    options.update(lavcopts=lavcopts)
    options.update(ffourcc='XVID')
    return options


def get_video_copy_options():
    return HyphenArgs(ovc='copy', ffourcc='XVID')


def get_audio_copy_options():
    return HyphenArgs(
        oac='copy',
    )


def get_mp3_options():
    return HyphenArgs(
        oac='mp3lame',
        lameopts=ColonDelimitedArgs(
            abr=None,
            br='96',
            vol='6',
        ),
    )


def encode_dvd():
    """
    %prog <dvd_source>

    Encode a DVD where the source is a DVD drive or RIP directory of a DVD.
    """
    parser = argparse.ArgumentParser(usage=trim(encode_dvd.__doc__))
    # parser.add_argument(
    #     '-t', '--titles',
    #     'enter the title or titles to process (i.e. 1 or 1,5 or 1-5)',
    #     default='')
    parser.add_argument(
        '-t', '--title', help='enter the dvd title number to process', default=''
    )
    parser.add_argument('-s', '--subtitle', help='enter the subtitle ID')
    parser.add_argument(
        '--test', help='just encode one chapter', default=False, action='store_true'
    )
    parser.add_argument(
        '-l',
        '--log-level',
        help='log level (debug, info, warning, error)',
        default='info',
    )
    parser.add_argument('device', nargs='?')
    args = parser.parse_args()

    logging.basicConfig(level=getattr(logging, args.log_level.upper()))

    command = MEncoderCommand()

    device = args.device or input('enter device> ')

    print('device is', device)
    command.set_device(device)

    videos_path = join(os.environ['PUBLIC'], 'Videos', 'Movies')

    default_title = infer_name(device)
    title_prompt = 'Enter output filename [%s]> ' % default_title
    user_title = input(title_prompt) or default_title

    filename = '%(user_title)s.avi' % vars()
    target = os.path.join(videos_path, user_title)
    output_filename = os.path.join(videos_path, filename)

    command['o'] = output_filename

    dvd_title = args.title
    command.source = ['dvd://%(dvd_title)s' % vars()]
    if args.test:
        command['chapter'] = '2-2'

    command.audio_options = get_audio_copy_options()
    command.audio_args.update(aid='128')

    crop = cropdetect.get_crop(device, dvd_title)
    log.info('crop is %s', crop)
    command.video_filter = HyphenArgs(
        sws='2',
        vf=ColonDelimitedArgs(crop=crop),
    )

    command.video_options = get_mpeg4_options()

    if args.subtitle:
        command['sid'] = args.subtitle
        command['vobsubout'] = target
        command['vobsuboutindex'] = '0'

    assert not os.path.exists(command.other_options['o']), (
        'Output file %s already exists' % command.other_options['o']
    )

    errors = open(os.devnull, 'w')
    two_pass_handler = MultiPassHandler(command)
    for _pass in two_pass_handler:
        _pass_args = tuple(_pass.get_args())
        log.debug('executing with %s', _pass_args)
        proc = subprocess.Popen(_pass_args, stderr=errors)
        proc.wait()


def re_encode(file, video_options, audio_options):
    command = MEncoderCommand()
    fn, ext = os.path.splitext(file)
    dest = ''.join((fn, ' (fixed)', ext))
    command['o'] = dest
    command.source = [file]
    command.audio_options = audio_options
    command.video_options = video_options
    errors = open(os.devnull, 'w')
    print('executing with', tuple(command.get_args()))
    proc = subprocess.Popen(command.get_args(), stderr=errors)
    if proc.wait() == 0:
        print('success')


def transcode():
    """
    %prog <source_file>

    Transcode a video by copying the video, but encoding the audio
    into mp3 format.
    """
    parser = argparse.ArgumentParser(usage=trim(fix_fourcc.__doc__))
    parser.add_argument('source')
    args = parser.parse_args()

    re_encode(args.source, get_video_copy_options(), get_mp3_options())


def fix_fourcc():
    """
    %prog <source_file>

    Re-encode the video, but change the ffourcc to XVID
    """
    parser = argparse.ArgumentParser(usage=trim(fix_fourcc.__doc__))
    parser.add_argument('source')
    options, args = parser.parse_args()

    re_encode(args.source, get_video_copy_options(), get_audio_copy_options())


def rip_subtitles():
    """
    %prog <dvd_source>
    """
    logging.basicConfig(level=logging.INFO)

    parser = argparse.ArgumentParser(usage=trim(encode_dvd.__doc__))
    parser.add_argument(
        '-t', '--title', help='enter the dvd title number to process', default=''
    )
    parser.add_argument('-s', '--subtitle', help='enter the subtitle ID')
    parser.add_argument('device', nargs='?')
    args = parser.parse_args()

    command = MEncoderCommand()

    device = args.device or input('enter device> ')

    print('device is', device)
    command.set_device(device)

    videos_path = join(os.environ['PUBLIC'], 'Videos', 'Movies')

    default_title = infer_name(device)
    title_prompt = 'Enter output filename [%s]> ' % default_title
    user_title = input(title_prompt) or default_title
    target = os.path.join(videos_path, user_title)

    command.source = ['dvd://%(title)s' % vars(args)]

    command['o'] = os.devnull

    command.audio_options = HyphenArgs(nosound=None)
    command.video_options = HyphenArgs(ovc='frameno')

    command['sid'] = args.subtitle or '0'

    command['vobsubout'] = target
    command['vobsuboutindex'] = command['sid']
    # command['vobsuboutid'] = 'en'

    command = tuple(command.get_args())
    errors = open(os.devnull, 'w')
    proc = subprocess.Popen(command, stderr=errors)
    proc.wait()


def update_anydvd():
    import jaraco.net.http

    url = 'http://static.slysoft.com/SetupAnyDVD.exe'
    file = jaraco.net.http.get_url(url)
    subprocess.check_call([file, '-S'])
    os.remove(file)
    os.remove(r'C:\Users\Public\Desktop\AnyDVD.lnk')



"""
Concatenate SRT files based on the length of associated media files.
"""

import contextlib
import functools
import re
import subprocess
import itertools
import datetime
import sys
from more_itertools import islice_extended as islice

import autocommand
import path
from . import srt
from tempora import parse_timedelta


@contextlib.contextmanager
def write_names_to_file(names):
    with path.TempDir() as dir:
        file = dir.joinpath('names.txt')
        lines = (f"file '{name}'" for name in names)
        file.write_text('\n'.join(lines))
        yield file


def get_duration(media_file):
    """
    https://superuser.com/questions/650291/how-to-get-video-duration-in-seconds
    """
    cmd = [
        'ffprobe',
        '-v',
        'error',
        '-show_entries',
        'format=duration',
        '-of',
        'default=noprint_wrappers=1:nokey=1',
        media_file,
    ]
    return parse_timedelta(subprocess.check_output(cmd, text=True))


def accumulate(increments):
    """
    >>> list(accumulate([1, 2, 3]))
    [1, 3, 6]
    """
    items_ = iter(increments)
    value = next(items_)
    yield value
    for item in items_:
        value += item
        yield value


@autocommand.autocommand(__name__)
def main(*files, include='.*', outfile=sys.stdout):
    matcher = functools.partial(re.search, include)
    matched = map(path.Path, filter(matcher, files))
    srt_files = list(matched)
    media_files = [srt_file.with_suffix('.mp4') for srt_file in srt_files]
    durations = map(get_duration, media_files)
    incr_additions = itertools.chain((datetime.timedelta(),), islice(durations, -1))
    additions = accumulate(incr_additions)
    with autocommand.smart_open(outfile, mode='w') as strm:
        for srt_file, addition in zip(srt_files, additions):
            entries = srt.SubEntry.get_entries(srt_file)
            for entry in entries:
                print(entry + addition, file=strm)



import os
import subprocess
import argparse
import re
import datetime
import threading
import platform
import sys
import ctypes

import path
import jaraco.path
from jaraco.ui import menu
from jaraco.functools import call_aside

try:
    from jaraco.windows.power import no_sleep
except ImportError:
    from jaraco.context import null as no_sleep
from more_itertools.recipes import consume

from . import dvd
from . import config


@call_aside
def add_hidden_method():
    """
    >>> path.Path('.').is_hidden()
    False
    """
    path.Path.is_hidden = jaraco.path.is_hidden


def source_is_high_def():
    blueray_dir = os.path.join(dvd.get_source(), 'BDMV')
    return os.path.isdir(blueray_dir)


def get_handbrake_cmd():
    quality = 22 if source_is_high_def() else 20
    return [
        'HandBrakeCLI',
        '-i',
        dvd.get_source(),
        '--subtitle',
        'scan',
        '--subtitle-forced',
        '--native-language',
        'eng',
        '--encoder',
        'x264',
        '--quality',
        str(quality),
        # '--large-file',
    ]


class TitleInfo:
    def __init__(self, title_no, title, episode, root, ext):
        self.__dict__.update(vars())
        del self.self

    def __iter__(self):
        "Return the parameters to handbrake to rip for this title"
        return iter(['-t', str(self.title_no), '-o', self.root / self.filename])

    @property
    def filename(self):
        return '{episode:02} - {title}{ext}'.format(**vars(self))

    def __str__(self):
        return self.title


def get_titles(root):
    title_durations()
    title_no = eval(input('enter starting DVD title number> '))
    visible_files = sorted(file for file in root.files() if not file.is_hidden())
    if visible_files:
        last_file = visible_files[-1].basename()
        print('last file is', last_file)
        last_episode = int(re.match(r'\d+', last_file).group(0)) + 1
    else:
        last_episode = 1
    prompt = f'enter starting episode [{last_episode}]> '
    episode = eval(input(prompt) or 'None') or last_episode
    ext = '.mp4'
    while True:
        title = input('enter title> ')
        if not title:
            return
        yield TitleInfo(title_no, title, episode, root, ext)
        title_no += 1
        episode += 1


def quick_brake():
    name = dvd.infer_name()
    title = input(f"Movie title ({name})> ") or name
    config.movies_root.isdir() or config.movies_root.makedirs()
    init_environment()
    dest = config.movies_root / title + '.mp4'
    cmd = get_handbrake_cmd() + [
        '--main-feature',
        '-o',
        dest,
    ]
    with no_sleep():
        subprocess.Popen(cmd).wait()


def find_root():
    root = config.tv_root
    choices = [showdir.basename() for showdir in root.dirs()]
    show = menu.Menu(choices).get_choice('Choose show (blank for new)> ')
    if not show:
        show = input('Show name> ')
    show_dir = root / show
    show_dir.makedirs_p()
    choices = [seasondir.basename() for seasondir in show_dir.dirs()]
    season = menu.Menu(choices).get_choice('Choose season (blank for new)> ')
    if not season:
        season = 'Season %d' % eval(input('Season Number> '))
    season_dir = show_dir / season
    season_dir.makedirs_p()
    return season_dir


def get_starts(stream, limit):
    """
    Read lines from a stream, but only read the first `limit` bytes of each
    line (in order to read text from an incomplete line).
    """
    while True:
        res = stream.read(limit)
        if not res:
            return
        yield res.decode('utf-8')
        stream.readline()


def two_stage_encode(args):
    """
    Handbrake does a rip/encode and then multiplexes (muxes) the audio
    with the encoded video. This function watches handbrake for when the
    multiplexing begins and then returns a thread that will complete the
    multiplexing so that another rip/encode job can proceed.
    """
    proc = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    for start in get_starts(proc.stdout, 7):
        if 'Muxing' in start:
            # start a thread to finish the process
            print('Muxing...')
            t = threading.Thread(target=proc.communicate)
            t.start()
            return t


def multibrake():
    root = find_root()
    parser = argparse.ArgumentParser()
    parser.add_argument('rest', nargs=argparse.REMAINDER)
    rest = parser.parse_args().rest
    threads = []
    with no_sleep():
        for title in list(get_titles(root)):
            args = get_handbrake_cmd() + rest + list(title)
            print('ripping', title)
            threads.append(two_stage_encode(args))
        [t.join() for t in threads if t]


def parse_duration(str):
    hours, minutes, seconds = map(int, str.split(':'))
    return datetime.timedelta(hours=hours, minutes=minutes, seconds=seconds)


def _link_to_title(lines):
    res = []
    for line in lines:
        title = re.match(r'\+ title (?P<title>\d+):', line)
        if title:
            res.append(dict(title=title.group('title')))
            continue
        m = re.match(r'  \+ (?P<key>.*): (?P<value>.*)', line)
        d = m.groupdict()
        key, value = d['key'], d['value']
        if key == 'duration':
            value = parse_duration(value)
        res[-1].update({key: value})
    return res


def more_than_ten_min(title):
    return 'duration' in title and title['duration'] > datetime.timedelta(minutes=10)


def title_durations():
    cmd = get_handbrake_cmd() + ['-t', '0']
    print('scanning...')
    output = subprocess.check_output(cmd, stderr=subprocess.STDOUT)
    output = output.decode('utf-8')
    lines = [
        line
        for line in output.splitlines()
        if '+ title' in line or '+ duration:' in line
    ]
    lines = _link_to_title(lines)
    lines = filter(more_than_ten_min, lines)
    print("Title durations:")
    [print(line['title'], line['duration']) for line in lines]


def init_environment():
    if platform.system() != 'Darwin':
        return
    libs = path.Path('/Applications/MakeMKV.app/Contents/lib')
    lib = next(iter(libs.glob('libmmbd*.dylib')), None)
    if not lib:
        print("Need to install MakeMKV", file=sys.stderr)
        raise SystemExit(1)
    root = path.Path('~/lib').expanduser()
    root.makedirs_p()
    link_names = 'libaacs.dylib', 'libbdplus.dylib'
    consume(
        lib.symlink(link)
        for link in map(root.joinpath, link_names)
        if not link.exists()
    )

    try:
        ctypes.CDLL('/usr/local/lib/libdvdcss.2.dylib')
    except Exception:
        msg = "Need libdvdcss (brew install libdvdcss)"
        print(msg, file=sys.stderr)
        raise SystemExit(1)



#!/usr/bin/python
'''
routines for acquiring dvd details, based on dvdinfo.py (many thanks
to Sybren Stüvel, http://stuvel.eu/dvdinfo).
'''

import re
import sys
import datetime
import argparse
from itertools import count
from subprocess import Popen, PIPE, STDOUT
from typing import Set

try:
    from importlib import metadata  # type: ignore
except ImportError:
    import importlib_metadata as metadata  # type: ignore


def banner():
    """
    Display the banner

    >>> banner()
    ==================================================
    jaraco.media version ...
    Jason R. Coombs <jaraco@jaraco.com>
    https://github.com/jaraco/jaraco.media
    ==================================================
    <BLANKLINE>
    """
    md = metadata.metadata('jaraco.media')

    print(50 * '=')
    print(f'{md["Name"]} version {md["Version"]}')
    print(f'{md["Author"]} <{md["Author-Email"]}>')
    print(f'{md["Home-page"]}')
    print(50 * '=')
    print()


class MetaTitleParser(type):
    """
    A metaclass for title parsers that keeps track of all of them.
    """

    _all_parsers: 'Set[MetaTitleParser]' = set()

    def __init__(cls, name, bases, attrs):
        cls._all_parsers.add(cls)
        # remove any base classes
        cls._all_parsers -= set(bases)


class TitleParser(metaclass=MetaTitleParser):
    def __init__(self, info):
        # info is the object that stores title info
        self.info = info
        self.pattern = re.compile(self.pattern)

    def __call__(self, line):
        self.parse(line)

    def parse(self, line):
        match = self.pattern.search(line)
        match and self.handle(match)

    @classmethod
    def create_all(cls, info):
        "Create all of the title parsers associated with this info"
        return [parser(info) for parser in cls._all_parsers]


class MaxTitlesParser(TitleParser):
    pattern = r'^There are (?P<max_titles>\d+) titles'

    def handle(self, match):
        self.info['max_titles'] = int(match.groupdict()['max_titles'])


class ChaptersParser(TitleParser):
    pattern = r'^ID_DVD_TITLE_(?P<title>\d+)_CHAPTERS=(?P<chapters>\d+)'

    def handle(self, match):
        title, chapters = map(int, map(match.groupdict().get, ['title', 'chapters']))
        if title == self.info['number']:
            self.info['chapters'] = chapters


class AudioParser(TitleParser):
    pattern = (
        r'^audio stream: (?P<stream>\d+) format: (?P<format>.+) '
        r'language: (?P<language>.+) aid: (?P<aid>\d+)'
    )

    def handle(self, match):
        '''Parse a single audio-channel line'''
        d = match.groupdict()
        d['aid'] = int(d['aid'])
        self.info['audiotracks'][d['aid']] = d


class SubtitleParser(TitleParser):
    pattern = r'^subtitle \(\s*sid\s*\): (?P<sid>\d+) language: (?P<language>.*)'

    def handle(self, match):
        '''Parse a single subtitle-channel line'''
        d = match.groupdict()
        d['sid'] = int(d['sid'])
        self.info['subtitles'].append(d)


class NaviParser(TitleParser):
    pattern = 'Found NAVI packet!'

    def handle(self, match):
        self.info['navi_count'] += 1


class LengthParser(TitleParser):
    pattern = r'ID_LENGTH=(?P<length>[\d.]+)'

    def handle(self, match):
        length = float(match.groupdict()['length'])
        length = datetime.timedelta(seconds=length)
        self.info['length'] = length


class TitleInfo(dict):
    def __init__(self, *args, **kwargs):
        self.update(
            max_titles=0, chapters=0, audiotracks={}, subtitles=[], navi_count=0
        )
        dict.__init__(self, *args, **kwargs)

    def __str__(self):
        buffer = []
        buffer.append('Number: %(number)s' % self)
        buffer.append('Title length: %(length)s' % self)
        buffer.append('Chapters: %(chapters)s' % self)
        buffer.append('Audio tracks:')
        aud_fmt = '\taid=%(aid)3i lang=%(language)s fmt=%(format)s'

        def fmt_aud_info(i):
            return aud_fmt % i

        buffer.extend(map(fmt_aud_info, self['audiotracks'].values()))
        buffer.append('Subtitles:')
        sub_fmt = '\tsid=%(sid)3i lang=%(language)s'

        def fmt_sub_info(i):
            return sub_fmt % i

        buffer.extend(map(fmt_sub_info, self['subtitles']))
        return '\n'.join(buffer)

    def has_audio(self):
        return self['audiotracks']


def title_info(device, title):
    """Get title information about a single title.

    expects device to be available globally.

    Returns a TitleInfo object
    """

    # need at least two -v to get "Found NAVI packet"
    mpcmd = (
        'mplayer -v -v -v -identify -nosound -frames 0 '
        '-dvd-device %s dvd://%i -vo null'
    )

    cmd = mpcmd % (device, title)
    mplayer = Popen(cmd, stdout=PIPE, stderr=STDOUT)

    info = TitleInfo(number=title)

    parsers = TitleParser.create_all(info)

    for line in mplayer.stdout:
        for parser in parsers:
            parser(line)
        if info['navi_count'] > 100:
            break

    return info


def main():
    longest_title_info = None

    banner()

    parser = argparse.ArgumentParser(usage=__doc__)
    parser.add_argument(
        '-t', '--title', help='only search a specific title', type=int, default=0
    )
    parser.add_argument('-d', '--device', help='the device (default d:)', default='d:')
    args = parser.parse_args()

    if not args.title:
        titles = []
        max_title = '?'
        # Walk through all titles
        for title in count(1):
            if isinstance(max_title, int) and title > max_title:
                break
            sys.stdout.write('Reading title %i/%s   \r' % (title, max_title))
            sys.stdout.flush()

            info = title_info(args.device, title)
            titles.append(info)
            # Remember info about the title with the most chapters,
            # but only if it has audio tracks.
            if info['audiotracks'] and (
                longest_title_info is None
                or info['chapters'] > longest_title_info['chapters']
            ):
                longest_title_info = info

            max_title = info['max_titles']

        print('Done reading.            ')

        titles_with_audio = filter(TitleInfo.has_audio, titles)
        titles_with_audio.sort(key=lambda t: -t['chapters'])

        if not titles_with_audio:
            raise SystemExit("Unable to find any titles with audio on %s" % args.device)

        for title in titles_with_audio:
            print()
            print(title)
    else:
        print('Reading title:', args.title)
        # Get info about given title
        info = title_info(args.device, args.title)

        print(info)


if __name__ == '__main__':
    main()



import datetime
import itertools
import argparse
import sys

_sample_srt_entries = """
1
00:00:25,080 --> 00:00:29,580
jaraco.media.srt presents

2
00:00:29,590 --> 00:00:33,660
An SRT Subtitle Parser

3
00:00:38,370 --> 00:00:47,870
Copyright ©2009-2010 Jason R. Coombs
Licensed for redistribution under the MIT license.

""".lstrip()


class SubEntry:
    r"""
    An SRT subtitle entry parser and manager.
    To get the SubEntry entries from an SRT file, just call get_entries
    on a filename or file-like object.

    >>> import io
    >>> sample = io.StringIO(_sample_srt_entries)
    >>> entries = SubEntry.get_entries(sample)

    Entries is now a generator which will parse the entries as they're
    requested. If you have enough memory to load them all in, then you
    can load them into a list or tuple.
    >>> entries = list(entries)

    >>> len(entries)
    3
    >>> entries[0].text
    ['jaraco.media.srt presents\n']
    >>> entries[1].start.seconds
    29
    >>> entries[1].start.microseconds
    590000

    Then, you can adjust when the entries occur. For example, this adds
    one second to the start and end times for the last subtitle.
    >>> entries[2] += datetime.timedelta(seconds=1)

    Finally, you can easily re-serialize the entries.
    >>> str(entries[2])
    '3\n00:00:39,370 --> 00:00:48,870\nCopyright... MIT license.\n'
    """

    time_sep = ' --> '

    def __init__(self, index, start, stop, text):
        self.__dict__.update(vars())
        del self.__dict__['self']

    @staticmethod
    def get_entries(file):
        if isinstance(file, str):
            file = open(file)
        while True:
            try:
                yield SubEntry.get_entry(file)
            except StopIteration:
                return

    @staticmethod
    def get_entry(items):
        items = iter(items)
        index = int(next(items))
        start, stop = SubEntry.parse_span(next(items))

        def is_not_blank(s):
            return bool(s.strip())

        text = list(itertools.takewhile(is_not_blank, items))
        return SubEntry(index, start, stop, text)

    @staticmethod
    def parse_span(span_string):
        times = span_string.split(SubEntry.time_sep)
        return map(SubEntry.parse_time, times)

    @staticmethod
    def parse_time(time_string):
        hr, min, sec = time_string.split(':')
        sec, msec = sec.split(',')
        return datetime.timedelta(
            hours=int(hr),
            minutes=int(min),
            seconds=int(sec),
            milliseconds=int(msec),
        )

    def __sub__(self, difference):
        "Subtract a time difference from the start and stop values"
        start = self.start - difference
        stop = self.stop - difference
        entry = SubEntry(self.index, start, stop, self.text)
        return entry

    def __add__(self, addition):
        "Add a time value to the start and stop values"
        start = self.start + addition
        stop = self.stop + addition
        entry = SubEntry(self.index, start, stop, self.text)
        return entry

    def format_span(self):
        times = map(self.format_time, [self.start, self.stop])
        return self.time_sep.join(times)

    @staticmethod
    def format_time(time):
        # python makes us construct a datetime object in order to get the desired
        #  time object in order to use strftime

        # pick some abitrary date
        dt = datetime.datetime(2000, 1, 1)
        # add our timedelta
        dt += time
        # extract the time
        time = dt.time()
        time_string = time.strftime('%H:%M:%S,%f')
        # strip the microseconds
        time_string = time_string[:-3]
        return time_string

    def __str__(self):
        items = [
            str(self.index),
            self.format_span(),
            ''.join(self.text),
        ]
        return '\n'.join(items)

    @staticmethod
    def write_entries(file, entries):
        if isinstance(file, str):
            file = open(file, 'w')
        for entry in entries:
            file.write(str(entry))
            file.write('\n')

    @staticmethod
    def adjust_entries(entries, adjustment):
        for entry in entries:
            yield entry + adjustment


class AdjustCommand:
    @classmethod
    def get_args(cls):
        parser = argparse.ArgumentParser()
        parser.add_argument(
            'delta', type=lambda s: datetime.timedelta(seconds=float(s))
        )
        parser.add_argument('-s', '--start-entry', default=0, type=int)
        parser.add_argument('-e', '--end-entry', type=int)
        return parser.parse_args()

    @classmethod
    def run(cls):
        args = cls.get_args()
        entries = SubEntry.get_entries(sys.stdin)
        start_entries = itertools.islice(entries, args.start_entry)
        if args.end_entry:
            # since start_entries may already have been consumed, adjust the
            #  end to account for those items.
            args.end_entry = args.end_entry - args.start_entry
        adjust_entries = itertools.islice(entries, args.end_entry)
        adjust_entries = (entry + args.delta for entry in adjust_entries)
        rest = entries

        adjusted_entries = itertools.chain(start_entries, adjust_entries, rest)

        SubEntry.write_entries(sys.stdout, adjusted_entries)



"""
jaraco.media assumes videos are stored under $root/Movies and $root/TV
where root is ~/Videos on Windows, ~ on other platforms, and can be
overridden by defining
VIDEOS_ROOT environment variable.
"""

import os
import platform

import path


def get_media_root():
    default = dict(
        Windows='~/Videos',
    ).get(platform.system(), '~')
    root = path.Path(os.environ.get('VIDEOS_ROOT', default))
    return root.expanduser()


movies_root = get_media_root() / 'Movies'
tv_root = get_media_root() / 'TV'



"""
Use ffmpeg to concatenate mpeg files.
"""

import contextlib
import functools
import re
import subprocess

import autocommand
import path


@contextlib.contextmanager
def write_names_to_file(names):
    with path.TempDir() as dir:
        file = dir.joinpath('names.txt')
        lines = (f"file '{name}'" for name in names)
        file.write_text('\n'.join(lines))
        yield file


@autocommand.autocommand(__name__)
def main(*files, out_name='out.mp4', include='.*'):
    matcher = functools.partial(re.search, include)
    matched = filter(matcher, files)
    with write_names_to_file(matched) as names_file:
        cmd = [
            'ffmpeg',
            '-f',
            'concat',
            # '-safe 0' required if input filenames are absolute :/
            '-safe',
            '0',
            '-i',
            names_file,
            '-c',
            'copy',
            out_name,
        ]
        subprocess.check_call(cmd)



import subprocess
from importlib_resources import files

import path

from jaraco.media import handbrake, dvd


def test_title_durations(monkeypatch):
    mocked = files(__package__).joinpath('multi title output.txt').read_bytes()
    monkeypatch.setattr(subprocess, 'check_output', lambda *args, **kw: mocked)
    monkeypatch.setattr(dvd, 'get_source', lambda: path.Path('/'))
    handbrake.title_durations()



import abc
from collections import OrderedDict

__all__ = ['DelimitedArgs', 'HyphenArgs', 'ColonDelimitedArgs']


class DelimitedArgs(OrderedDict):
    value_join = '='

    @abc.abstractproperty
    def delimiter(self):
        "The delimited between arguments"

    def __str__(self):
        return self.delimiter.join(self.get_args())

    arg_items = OrderedDict.items

    def get_args(self):
        return [
            self.value_join.join(item for item in arg if item)
            for arg in self.arg_items()
        ]


class HyphenArgs(DelimitedArgs):
    """
    Construct args suitable for unix-style command lines.

    e.g. -flag
    >>> print(HyphenArgs({'flag':None}))
    -flag

    e.g. -filename myfile.txt
    >>> print(HyphenArgs(filename='myfile.txt'))
    -filename myfile.txt

    >>> args = HyphenArgs([('a','a'), ('b','b')])
    >>> args_copy = args.copy()
    >>> print(args_copy)
    -a a -b b
    >>> print(HyphenArgs([('a', '1'), ('b', None)]))
    -a 1 -b
    """

    value_join = ' '
    delimiter = ' '

    @staticmethod
    def add_hyphen(value):
        return '-%s' % value

    def arg_items(self):
        return zip(self.hyphenated_keys(), self.values())

    def hyphenated_keys(self):
        return map(self.add_hyphen, super(self.__class__, self).keys())


class ColonDelimitedArgs(DelimitedArgs):
    """
    >>> args = ColonDelimitedArgs((('x', '3'), ('y', '4')))
    >>> print(args)
    x=3:y=4
    """

    delimiter = ':'
